package main

import (
	protocolJson "consumerDemo-json_go/json"
	"consumerDemo-json_go/sasl_sarama"
	"context"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"github.com/Shopify/sarama"
	"strconv"
	"strings"
)

var (
	partitionConsumers = make(map[int32]*partitionMsgConsumer)
	brokerList         = flag.String("brokers", "", "REQUIRED: The comma separated list of brokers in the Kafka cluster")
	topic              = flag.String("topic", "", "REQUIRED: the topic to consume")
	group              = flag.String("group", "", "REQUIRED: the consumer group")
	user               = flag.String("user", "", "the user name")
	password           = flag.String("password", "", "the password")
	trans2sql          = flag.Bool("trans2sql", false, "trans message to sql or not")
)

func main() {
	flag.Parse()
	config := sarama.NewConfig()
	config.Version = sarama.V0_11_0_0
	config.Consumer.Return.Errors = true
	config.Consumer.Offsets.Initial = sarama.OffsetOldest
	config.Consumer.Fetch.Default = 20 << 20
	if *user != "" {
		config.Net.SASL.Enable = true
		config.Net.SASL.User = *user
		config.Net.SASL.Password = *password
		config.Net.SASL.Handshake = true
		config.Net.SASL.SCRAMClientGeneratorFunc = func() sarama.SCRAMClient { return &sasl_sarama.XDGSCRAMClient{HashGeneratorFcn: sasl_sarama.SHA512} }
		config.Net.SASL.Mechanism = sarama.SASLTypeSCRAMSHA512
	}

	// Start with a client
	client, err := sarama.NewClient(strings.Split(*brokerList, ","), config)
	if err != nil {
		panic(err)
	}
	defer func() { _ = client.Close() }()

	// Start a new consumer group
	group, err := sarama.NewConsumerGroupFromClient(*group, client)
	if err != nil {
		panic(err)
	}
	defer func() { _ = group.Close() }()

	// Track errors
	go func() {
		for err := range group.Errors() {
			fmt.Println("ERROR", err)
		}
	}()

	// Iterate over consumer sessions.
	ctx := context.Background()
	for {
		handler := demoConsumerGroupHandler{}
		// 开始消费
		err := group.Consume(ctx, []string{*topic}, handler)
		if err != nil {
			panic(err)
		}
	}
}

type demoConsumerGroupHandler struct {
}

func (demoConsumerGroupHandler) Setup(sess sarama.ConsumerGroupSession) error {
	setup := make(map[int32]bool)
	for _, part := range sess.Claims()[*topic] {
		if _, ok := partitionConsumers[part]; !ok {
			partitionConsumers[part] = &partitionMsgConsumer{}
		}

		setup[part] = true
	}

	var invalid []int32
	for p := range partitionConsumers {
		if _, ok := setup[p]; !ok {
			invalid = append(invalid, p)
		}
	}

	for _, p := range invalid {
		delete(partitionConsumers, p)
	}

	return nil
}

func (demoConsumerGroupHandler) Cleanup(_ sarama.ConsumerGroupSession) error { return nil }

func (h demoConsumerGroupHandler) ConsumeClaim(sess sarama.ConsumerGroupSession, claim sarama.ConsumerGroupClaim) error {
	for msg := range claim.Messages() {
		if err := partitionConsumers[msg.Partition].consume(sess, msg); nil != err {
			return err
		}
	}
	return nil
}

type partitionMsgConsumer struct {
	lastPartitionSeq  uint64
	lastPartitionSeq2 uint64
	largeRecord       *protocolJson.FlatRecord
}

func (pmc *partitionMsgConsumer) consume(sess sarama.ConsumerGroupSession, msg *sarama.ConsumerMessage) error {
	//for _, header := range msg.Headers {
	//	fmt.Println(fmt.Sprintf("key: %s, value: %s", string(header.Key), string(header.Value)))
	//}
	// Each message has a sequence id in the partition, this id is continuous monotonic increasing,
	// these messages are duplicated if the id rotate;
	// There is unexpected message lost in kafka if the id is discontinuous,
	// have to reset the subscribe and client when it happened.
	// ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
	// ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
	// 重复解析部分binlog导致的数据重复投递问题。
	ps2, err := getPartitionSeq2(msg)
	var validPs uint64
	if err == nil {
		if pmc.lastPartitionSeq2 > 0 {
			if ps2 < pmc.lastPartitionSeq2 {
				fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
					msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
				return nil
			} else if ps2 == pmc.lastPartitionSeq2 { // 可能是重复消息也可能是checkpoint或heartbeat消息
				if !isCheckpoint(msg) && !isHeartbeat(msg) {
					fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
						msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
					return nil
				}
			} else {
				if ps2 != pmc.lastPartitionSeq2+1 {
					errMsg := fmt.Sprintf("the PartitionSeq2 is discontinuous, last: %v, current: %v, "+
						"partition:%v, offset:%v", pmc.lastPartitionSeq2, ps2, msg.Partition, msg.Offset)
					panic(errMsg)
				}
			}
		}
		validPs = ps2
		pmc.lastPartitionSeq2 = validPs
	} else {
		ps, err := getPartitionSeq(msg)
		if err != PsNotExistError {
			if ps <= pmc.lastPartitionSeq {
				fmt.Printf("duplicated msg: %+v\n", msg)
				return nil
			} else if ps != pmc.lastPartitionSeq+1 && pmc.lastPartitionSeq != 0 {
				errMsg := fmt.Sprintf("the PartitionSeq is discontinuous, last: %v, current: %v, "+
					"partition:%v, offset:%v", pmc.lastPartitionSeq, ps, msg.Partition, msg.Offset)
				panic(errMsg)
			}
			validPs = ps
			pmc.lastPartitionSeq2 = validPs
		}
	}

	record := &protocolJson.FlatRecord{}
	// 反序列化
	err = json.Unmarshal(msg.Value, record)
	if err != nil {
		errMsg := fmt.Sprintf("Unmarshal error: %v", err.Error())
		panic(errMsg)
	}

	// version 2 需要考虑大消息合并，生产端对于超过10M的大消息做了拆分
	if record.Version == 2 && record.Total > 1 {
		fmt.Printf("merging large record, partition:%v, offset:%v, partitionSeq:%v, Total:%v, Index:%v, GTID:%v\n",
			msg.Partition, msg.Offset, validPs, record.Total, record.Index, record.Gtid)
		if record.Index == 0 {
			pmc.largeRecord = record
		} else {
			if record.OldColumnsBytes != nil {
				oldCols0 := pmc.largeRecord.OldColumnsBytes
				oldCols1 := record.OldColumnsBytes
				pmc.largeRecord.OldColumnsBytes = append(oldCols0, oldCols1...)
			}
			if record.NewColumnsBytes != nil {
				newCols0 := pmc.largeRecord.NewColumnsBytes
				newCols1 := record.NewColumnsBytes
				pmc.largeRecord.NewColumnsBytes = append(newCols0, newCols1...)
			}
		}
		if record.Index < record.Total-1 {
			return nil
		}
		if pmc.largeRecord.OldColumnsBytes != nil {
			oldColumns := new([]interface{})
			oldColumnsBytes := pmc.largeRecord.OldColumnsBytes
			fmt.Printf("largeRecord.oldColumnsBytes length: %v\n", len(oldColumnsBytes))
			err := json.Unmarshal(oldColumnsBytes, oldColumns)
			if err != nil {
				errMsg := fmt.Sprintf("Unmarshal oldColumnsBytes error: %v", err.Error())
				panic(errMsg)
			}
			pmc.largeRecord.OldColumns = *oldColumns
			pmc.largeRecord.OldColumnsBytes = nil
		}
		if pmc.largeRecord.NewColumnsBytes != nil {
			newColumns := new([]interface{})
			newColumnsBytes := pmc.largeRecord.NewColumnsBytes
			fmt.Printf("largeRecord.newColumnsBytes length: %v\n", len(newColumnsBytes))
			err := json.Unmarshal(newColumnsBytes, newColumns)
			if err != nil {
				errMsg := fmt.Sprintf("Unmarshal newColumnsBytes error: %v", err.Error())
				panic(errMsg)
			}
			pmc.largeRecord.NewColumns = *newColumns
			pmc.largeRecord.NewColumnsBytes = nil
		}
		record = pmc.largeRecord
		record.Index = -1
	}
	err = onEntry(msg.Partition, msg.Offset, validPs, record)
	if err != nil {
		errMsg := fmt.Sprintf("onEntry error: %v", err.Error())
		panic(errMsg)
	}
	//commit at checkpoint message
	if record.MessageType == protocolJson.CHECKPOINT {
		sess.MarkMessage(msg, "")
	}
	return nil
}

const partitionSeqKey = "ps"
const partitionSeq2Key = "ps2"

var PsNotExistError = errors.New(partitionSeqKey + " does not exists")

func getPartitionSeq(msg *sarama.ConsumerMessage) (uint64, error) {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeqKey {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq: " + string(m.Value))
			}
			return n, nil
		}
	}
	return 0, PsNotExistError
}

func getPartitionSeq2(msg *sarama.ConsumerMessage) (uint64, error) {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeq2Key {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq2: " + string(m.Value))
			}
			return n, nil
		}
	}
	return 0, errors.New(partitionSeq2Key + " does not exists")
}

func onEntry(partitionNum int32, offset int64, partitionSeq uint64, record *protocolJson.FlatRecord) error {
	b, err := json.Marshal(record)
	if err != nil {
		return err
	}
	fmt.Printf("-->[partition: %v, offset: %v, partitionSeq: %v]\n %s\n", partitionNum, offset, partitionSeq, string(b))
	if *trans2sql {
		return trans2SQL(record)
	}
	return nil
}

func trans2SQL(record *protocolJson.FlatRecord) error {
	switch record.MessageType {
	case protocolJson.HEARTBEAT:
		fmt.Println("heartbeat:", record.HeartbeatTimestamp)
	case protocolJson.CHECKPOINT:
		fmt.Println("Checkpoint: gtid=", record.SyncedGtid, ", file=", record.FileName, ", position=", record.Position)
	case protocolJson.BEGIN:
		fmt.Println("BEGIN")
	case protocolJson.COMMIT:
		fmt.Println("COMMIT")
	case protocolJson.ROLLBACK:
		fmt.Println("ROLLBACK")
	case protocolJson.DDL, protocolJson.INIT_DDL, protocolJson.STATEMENT:
		fmt.Println(record.Sql)
	case protocolJson.INSERT:
		if sql, err := transInsert(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case protocolJson.INIT_INSERT:
		if sql, err := transInitInsert(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case protocolJson.DELETE:
		if sql, err := transDelete(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case protocolJson.UPDATE:
		if sql, err := transUpdate(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	default:
		return errors.New("unsupported message type")
	}

	return nil
}

func transInitInsert(record *protocolJson.FlatRecord) (string, error) {
	if record.NewColumns == nil {
		return "", errors.New(fmt.Sprintf("empty data: %v", record))
	}

	switch array := record.NewColumns.(type) {
	case []interface{}:
		sql := &strings.Builder{}
		cols := make([]string, 0, len(record.Columns))
		for _, col := range record.Columns {
			cols = append(cols, col.Name)
		}
		sql.WriteString(fmt.Sprintf("INSERT INTO `%v`.`%v`(%s) VALUES ", record.SchemaName, record.TableName, strings.Join(cols, ",")))
		sql.WriteString("(")
		sep := ""
		for idx, value := range array {
			sql.WriteString(sep)
			sql.WriteString(fmt.Sprintf("%v", protocolJson.Decode(value, record.Columns[idx])))
			sep = ", "
		}
		sql.WriteString(")")
		return sql.String(), nil
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %T", record.NewColumns))
	}
}

func transInsert(record *protocolJson.FlatRecord) (string, error) {
	if record.NewColumns == nil {
		return "", errors.New(fmt.Sprintf("empty data: %v", record))
	}
	switch array := record.NewColumns.(type) {
	case []interface{}:
		sql := &strings.Builder{}
		sql.WriteString(fmt.Sprintf("INSERT INTO `%v`.`%v` VALUES ", record.SchemaName, record.TableName))
		sql.WriteString("(")
		sep := ""
		for idx, value := range array {
			sql.WriteString(sep)
			sql.WriteString(fmt.Sprintf("%v", protocolJson.Decode(value, record.Columns[idx])))
			sep = ", "
		}
		sql.WriteString(")")
		return sql.String(), nil
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %T", record.NewColumns))
	}
}

func transDelete(record *protocolJson.FlatRecord) (string, error) {
	if record.OldColumns == nil {
		return "", errors.New(fmt.Sprintf("empty data: %v", record))
	}
	//fmt.Printf("OldColumns type: %T\n", record.OldColumns)
	switch array := record.OldColumns.(type) {
	case []interface{}:
		sql := &strings.Builder{}
		sql.WriteString(fmt.Sprintf("DELETE FROM `%v`.`%v` WHERE ", record.SchemaName, record.TableName))
		columns := record.Columns
		sep := ""
		for idx, value := range array {
			d := protocolJson.Decode(value, columns[idx])
			if "DEFAULT" == d || "NULL" == d {
				continue
			}

			sql.WriteString(sep)
			sql.WriteString(fmt.Sprintf("`%v` = %v", columns[idx].Name, d))
			sep = " AND "
		}

		sql.WriteString(" LIMIT 1")
		return sql.String(), nil
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %T", record.OldColumns))
	}
}

func transUpdate(record *protocolJson.FlatRecord) (string, error) {
	if record.NewColumns == nil || record.OldColumns == nil {
		return "", errors.New(fmt.Sprintf("empty data: %v", record))
	}
	switch array := record.NewColumns.(type) {
	case []interface{}:
		sql := &strings.Builder{}
		sql.WriteString(fmt.Sprintf("UPDATE `%v`.`%v` SET ", record.SchemaName, record.TableName))
		columns := record.Columns
		sep := ""
		for idx, value := range array {
			if value == protocolJson.JSON_NA_VALUE {
				continue
			}

			sql.WriteString(sep)
			sql.WriteString(fmt.Sprintf("`%v` = %v", columns[idx].Name, protocolJson.Decode(value, record.Columns[idx])))
			sep = ", "
		}

		sql.WriteString(" WHERE ")
		sep = ""
		oldArray := record.OldColumns.([]interface{})
		for idx, value := range oldArray {
			d := protocolJson.Decode(value, columns[idx])
			if "DEFAULT" == d || "NULL" == d {
				continue
			}

			sql.WriteString(sep)
			sql.WriteString(fmt.Sprintf("`%v` = %v", columns[idx].Name, d))
			sep = " AND "
		}

		sql.WriteString(" LIMIT 1")
		return sql.String(), nil
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type, new columns %T, old columns: %T", record.NewColumns, record.OldColumns))
	}
}

func isCheckpoint(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Checkpoint" {
			return true
		}
	}
	return false
}

func isHeartbeat(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Heartbeat" {
			return true
		}
	}
	return false
}
